
/******************************************************************************
* MODULE     : relaxed.gen.cc
* DESCRIPTION: relaxed power series
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <Univariate/relaxed.gen.h>
#include <Univariate/Zealous/zealous.gen.cc>
#include "multiply.gen.cc"
#include "compose.gen.cc"

#module code_relaxed (C)
#import relaxed (C)
#import code_zealous (C)
#import code_relaxed_multiply (C)
#import code_relaxed_compose (C)

/******************************************************************************
* general routines
******************************************************************************/

C
series<C>::operator [] (int n) {
  if (n < rep->n) return rep->value[n];
  if (n >= N (rep->value)) rep->value= rep->value (0, 1 >? (n<<1));
  while (rep->n <= n) {
    rep->value[rep->n]= rep->next ();
    rep->n++;
  }
  // cout << ((tree) (*this)) << " [" << n << "] -> " << rep->value[n] << "\n";
  return rep->value[n];
}

zealous<C>
series<C>::operator () (int start, int end) {
  if (end <= rep->n) return rep->value (start, end);
  if (end > N (rep->value)) rep->value= rep->value (0, end);
  while (rep->n < end) {
    rep->value[rep->n]= rep->next ();
    rep->n++;
  }
  return rep->value (start, end);
}

series<C>::operator tree () {
  return (tree) (*rep);
}

/******************************************************************************
* implicit series
******************************************************************************/

void
implicit_series_rep<C>::set_precision (int terms) {
  ::series_rep<C>::set_precision (terms);
}

implicit_series<C>::implicit_series<C> (int r) {
  rep= new implicit_series_rep<C> (r);
  rep->ref_count++;
}

C&
implicit_series<C>::operator () (int n) {
  if (n>=rep->r) fatal_error ("Insufficient number of initial conditions",
			      "implicit_series<C>::operator []");
  return rep->value[n];
}

C
implicit_series<C>::operator [] (int n) {
  return ((series<C>) (*this)) [n];
}

zealous<C>
implicit_series<C>::operator () (int start, int end) {
  return ((series<C>) (*this)) (start, end);
}

struct auxiliary_series_rep<C>: public series_rep<C> {
  implicit_series<C> f;
  tree t;
  auxiliary_series_rep<C> (implicit_series<C> f2, tree t2):
    series_rep<C> (RELAXED), f(f2), t (t2) {}
  ~auxiliary_series_rep<C> () { f->eq= series<C>(); }
  C next () { return ((series<C>) f)[n]; }
  operator tree () { return t; }
};

series<C>
auxiliary_series<C> (implicit_series<C> f, tree t) {
  return new auxiliary_series_rep<C> (f, t);
}

/******************************************************************************
* constant and polynomial series
******************************************************************************/

struct polynomial_series_rep<C>: public series_rep<C> {
  polynomial_series_rep<C> (zealous<C> f):
    series_rep<C> (POLYNOMIAL) { value= f; n= N(f); }
  C next () { return ZERO<C>; }
  operator tree () { return tuple ("polynomial"); }
};

series<C>::series<C> (C c, int deg) {
  zealous<C> f (deg+1);
  f[deg]= c;
  rep= new polynomial_series_rep<C> (f);
  rep->ref_count++;
}

series<C>::series<C> (zealous<C> f) {
  rep= new polynomial_series_rep<C> (f);
  rep->ref_count++;
}

/******************************************************************************
* series whose coefficients are given by a function
******************************************************************************/

struct func_series_rep<C>: public series_rep<C> {
  C (*coeffs) (int);
  func_series_rep<C> (C (*coeffs2) (int)):
    series_rep<C> (RELAXED), coeffs (coeffs2) {}
  C next () { return coeffs(n); }
  operator tree () { return tuple ("func"); }
};

series<C>::series<C> (C (*coeffs) (int)) {
  rep= new func_series_rep<C> (coeffs);
  rep->ref_count++;
}

/******************************************************************************
* shifting
******************************************************************************/

struct shift_series_rep<C>: public series_rep<C> {
  series<C> f;
  int sh;
  shift_series_rep<C> (series<C> f2, int sh2):
    series_rep<C> (RELAXED), f(f2), sh(sh2) {}
  C next () {
    if ((n+sh)<0) return ZERO<C>;
    return f[n+sh];
  }
  operator tree () { return tuple (">>", f, tree (as_string (sh))); }
};

series<C>
operator >> (series<C> f, int n) {
  return new shift_series_rep<C> (f, n);
}

series<C>
operator << (series<C> f, int n) {
  return new shift_series_rep<C> (f, -n);
}

/******************************************************************************
* sums, differences and scalar products
******************************************************************************/

struct sum_series_rep<C>: public series_rep<C> {
  series<C> f, g;
  sum_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next () { return f[n]+ g[n]; }
  operator tree () { return tuple ("+", f, g); }
};

series<C>
operator + (series<C> f, series<C> g) {
  return new sum_series_rep<C> (f, g);
}

struct diff_series_rep<C>: public series_rep<C> {
  series<C> f, g;
  diff_series_rep<C> (series<C> f2, series<C> g2):
    series_rep<C> (RELAXED), f(f2), g(g2) {}
  C next () { return f[n]- g[n]; }
  operator tree () { return tuple ("-", f, g); }
};

series<C>
operator - (series<C> f, series<C> g) {
  return new diff_series_rep<C> (f, g);
}

struct scalar_mul_series_rep<C>: public series_rep<C> {
  C c; series<C> f;
  scalar_mul_series_rep<C> (C c2, series<C> f2):
    series_rep<C> (RELAXED), c (c2), f(f2) {}
  C next () { return c*f[n]; }
  operator tree () { return tuple ("*", c, f); }
};

series<C>
operator * (C c, series<C> f) {
  return new scalar_mul_series_rep<C> (c, f);
}

/******************************************************************************
* quotients
******************************************************************************/

struct aux_inv_series_rep<C>: public series_rep<C> {
  series<C> f;
  aux_inv_series_rep<C> (series<C> f2):
    series_rep<C> (RELAXED), f(f2) {}
  C next () { return -f[n+1]/f[0]; }
  operator tree () { return tuple ("aux-inv", f); }
};

series<C>
operator / (series<C> f, series<C> g) {
  implicit_series<C> h (1);
  h(0) = C(1)/g[0];
  h->eq= ((new aux_inv_series_rep<C> (g)) * h) << 1;
  return f * auxiliary_series<C> (h, tuple ("/", f, g));
}

/******************************************************************************
* differential calculus
******************************************************************************/

struct differentiated_series_rep<C>: public series_rep<C> {
  series<C> f;
  differentiated_series_rep<C> (series<C> f2):
    series_rep<C> (RELAXED), f(f2) {}
  C next () { return C(n+1)*f[n+1]; }
  operator tree () { return tuple ("d", f); }
};

series<C>
d (series<C> f) {
  return new differentiated_series_rep<C> (f);
}

#ifdef has_divisible<C>

struct scalar_div_series_rep<C>: public series_rep<C> {
  series<C> f; C c;
  scalar_div_series_rep<C> (series<C> f2, C c2):
    series_rep<C> (RELAXED), f(f2), c (c2) {}
  C next () { return f[n]/c; }
  operator tree () { return tuple ("/", f, c); }
};

series<C>
operator / (series<C> f, C c) {
  return new scalar_div_series_rep<C> (f, c);
}

struct integrated_series_rep<C>: public series_rep<C> {
  series<C> f;
  integrated_series_rep<C> (series<C> f2):
    series_rep<C> (RELAXED), f(f2) {}
  C next () { if (n==0) return 0; return f[n-1]/C(n); }
  operator tree () { return tuple ("I", f); }
};

series<C>
I (series<C> f) {
  return new integrated_series_rep<C> (f);
}

series<C>
exp (series<C> f) {
  implicit_series<C> g (1);
#ifdef has_special_functions<C>
  g(0) = exp (f[0]);
#else
  g(0) = ONE<C>;
#endif
  g->eq= I(g*d(f));
  return auxiliary_series<C> (g, tuple ("exp", f));
}

series<C>
log (series<C> f) {
#ifdef has_special_functions<C>
  return series<C>(log (f[0]))+ I(d(f)/f);
#else
  return I(d(f)/f);
#endif
}

#endif

/******************************************************************************
* output
******************************************************************************/

static int series<C>_precision= 10;

void
set_series<C>_precision (int n) {
  series<C>_precision= n;
}

ostream&
operator << (ostream& out, series<C> f) {
  return out << f(0, series<C>_precision);
}

#endmodule // code_relaxed (C)
